using Cairo;
using Gdk;
using Gtk;
using System;
using static System.Console;
using static System.Math;

using Color = Gdk.Color;
using Rectangle = Cairo.Rectangle;
using Window = Gtk.Window;

delegate void Notify();

// A model class for a game of Tic Tac Toe.
class Board {
  int[,] square = new int[3, 3];  // 0 = empty, 1 = player 1, 2 = player 2
  
  public int player = 1;   // whose turn it is
  public int winner = 0;   // who has won
  public int win_x, win_y, win_dx, win_dy;  // vector indicating winning squares
  
  public event Notify changed;  // fires whenever the board changes
  
  public int this[int x, int y] {
    get => square[x, y];
  }
  
  // Return true if the current player has three in a row,
  // starting at (x, y) and moving in direction (dx, dy).
  bool all(int x, int y, int dx, int dy) {
    for (int i = 0 ; i < 3 ; ++i) 
      if (square[x + i * dx, y + i * dy] != player)
        return false;
    
    win_x = x; win_y = y; win_dx = dx; win_dy = dy;
    return true;
  }

  // Return true if the current player has won.
  bool checkWin() {
    // check rows and columns
    for (int i = 0 ; i < 3 ; ++i)
      if (all(i, 0, 0, 1) || all(0, i, 1, 0))
        return true;
        
    // check diagonals
    return all(0, 0, 1, 1) || all(2, 0, -1, 1);
  }
  
  // Make the current player play at (x, y).  Return true if the move was legal.
  public bool move(int x, int y) {
    if (square[x, y] > 0)  // square is occupied
      return false;
      
    square[x, y] = player;
    if (checkWin())
      winner = player;
    else player = 3 - player;
    
    changed();
    return true;
  }
}

static class Util {
  public static Color parse(string name) {
    Color c = new Color(0, 0, 0);
    if (!Color.Parse(name, ref c))
      throw new Exception("bad color");
    return c;
  }
  
  // extension methods on Context
  
  public static void setColor(this Context c, Color color) {
    CairoHelper.SetSourceColor(c, color);
  }
  
  public static void drawLine(this Context c, double x1, double y1, double x2, double y2) {
    c.MoveTo(x1, y1);
    c.LineTo(x2, y2);
    c.Stroke();
  }
}

// A graphical interface for Tic Tac Toe.

class View : DrawingArea {
  const int SquareSize = 100;
  const int Margin = SquareSize / 2;
  
  Board board;
  
  void init() {
    board = new Board();
    QueueDraw();   // draw the initial position
    board.changed += QueueDraw;
  }
  
  public View() {
    init();
    AddEvents((int) EventMask.ButtonPressMask);  // ask to receive button press events
    ModifyBg(StateType.Normal, Util.parse("white"));
  }

  void drawX(Context c, Rectangle r) {
    c.drawLine(r.X, r.Y, r.X + r.Width, r.Y + r.Height);
    c.drawLine(r.X + r.Width, r.Y, r.X, r.Y + r.Height);
  }
  
  void drawO(Context c, Rectangle r) {
    c.Arc(r.X + r.Width / 2, r.Y + r.Height / 2, r.Width / 2, 0.0, 2 * PI);
    c.Stroke();
  }
  
  protected override bool OnButtonPressEvent(EventButton ev) {
    if (board.winner > 0)
      init();   // start a new game
    else {
      int x = ((int) (ev.X - Margin)) / SquareSize;
      int y = ((int) (ev.Y - Margin)) / SquareSize;
      board.move(x, y);
    }
    return true;
  }

  protected override bool OnExposeEvent(EventExpose ev) {
    Context c = CairoHelper.Create(GdkWindow);
    
    // Set a transformation matrix so that squares are 1 unit tall/wide, with a
    // margin around the drawing area.
    c.Translate(Margin, Margin);
    c.Scale(SquareSize, SquareSize);
    
    // Adjust the line width to compensate for the scaling tranformation.
    c.LineWidth /= SquareSize;
    
    if (board.winner > 0) {
      c.setColor(Util.parse("light green"));
      for (int i = 0 ; i < 3 ; ++i) {
        c.Rectangle(board.win_x + i * board.win_dx,
                    board.win_y + i * board.win_dy,
                    1, 1);
        c.Fill();
      }
      c.setColor(Util.parse("black"));
    }
    
    // Draw the tic tac toe grid.
    for (int i = 1 ; i <= 2 ; ++i) {
      c.drawLine(i, 0, i, 3);
      c.drawLine(0, i, 3, i);
    }
    
    // Draw the X's and O's.
    for (int x = 0 ; x < 3 ; ++x)
      for (int y = 0 ; y < 3 ; ++y) {
        Rectangle rect = new Rectangle(x + 0.1, y + 0.1, 0.8, 0.8);
        
        switch (board[x, y]) {
          case 1: drawX(c, rect); break;
          case 2: drawO(c, rect); break;
        }
      }
      
    c.GetTarget().Dispose();
    c.Dispose();
    return true;
  }
}

class Frame : Window {
  Frame() : base("tic tac toe") {
    Resize(400, 400);
    Add(new View());
  }
  
  protected override bool OnDeleteEvent(Event ev) {
    Application.Quit();
    return true;
  }
  
  static void Main() {
    Application.Init();
    new Frame().ShowAll();
    Application.Run();
  }
}
